////////////////////////////////////////////////////////////////////////////////
// Copyright  2003 Harry Denholm					  http://www.ishani.plus.com
// 
// header with basic functions to set up and use a channel to send
// text to the PC debug console via USB
////////////////////////////////////////////////////////////////////////////////

#ifndef HDD_USBDEBUGGING
#define HDD_USBDEBUGGING

//-----------------------------------------------------------------------------
// Include directives

#include "gpcomm.h"


//-----------------------------------------------------------------------------
// Macro definitions

#define	USBDEBUG_PKT_SIZE		(256)


//-----------------------------------------------------------------------------
// Enumerations & type definitions

static GPN_DESC gCommDesc;


//------------------------------------------------------------------------------
// Function declarations

////////////////////////////////////////////////////////////////////////////////
// USBDebuggingStart
//
// call this function to begin sending to the USB debug console on the PC
// sets up the GP32's communications library for USB device comms
void USBDebuggingStart(
	GPN_COMM *com)
{
	// setup comm system
	gm_memset(&gCommDesc, 0, sizeof(GPN_DESC));
	gCommDesc.port_kind = COMM_USB_D;
	gCommDesc.tr_mode = 1;
	gCommDesc.tr_buf_size = 2048;
	gCommDesc.sz_pkt = USBDEBUG_PKT_SIZE;
	gCommDesc.isr_comm_ram = NULL;
	gCommDesc.reserved1 = 0;
	gCommDesc.reserved2 = 0;
	
	// create communications channel
	GpCommCreate(&gCommDesc, com);
}


////////////////////////////////////////////////////////////////////////////////
// USBDebuggingSendText
//
// send the text buffer specified to the PC debug console
// maximum length of 256

//Torlus - PC->GP32 output management
int USBDebuggingSendText(
	GPN_COMM *com,
	char *inbuf, char *outbuf)
{		
	int len;
	
	// send the buffer
	com->comm_send((uchar*)inbuf, USBDEBUG_PKT_SIZE);
	
	// block until debugging console pings us back
	len = com->comm_recv_sync();
	
	// flush the input buffer
	com->comm_recv((uchar*)outbuf, USBDEBUG_PKT_SIZE);
	
	return len;
}

////////////////////////////////////////////////////////////////////////////////
// USBDebuggingStop
//
// called once debugging is finished
void USBDebuggingStop()
{
	GpCommDelete(&gCommDesc);
}

#endif // HDD_USBDEBUGGING


